@@ -27,7 +27,7 @@
  * @author Samuel Mimram, Julien Cristau, David Baelde
  */
 
-/* $Id: vorbis_stubs.c 2564 2006-07-04 12:57:08Z smimram $ */
+/* $Id: vorbis_stubs.c 2693 2006-10-09 17:04:50Z smimram $ */
 
 /*
 TODO:
@@ -43,6 +43,7 @@ TODO:
 #include <caml/memory.h>
 #include <caml/misc.h>
 #include <caml/mlvalues.h>
+#include <caml/signals.h>
 
 #include <vorbis/vorbisfile.h>
 #include <vorbis/vorbisenc.h>
@@ -148,9 +149,9 @@ inline static char* string_of_option(value v)
     return NULL;
 }
 
-#define copy_buffer(b,len,dst) \
-  dst = caml_alloc_string(len); \
-memmove(String_val(dst), b, len);
+#define copy_buffer(dst, b, len) \
+        dst = caml_alloc_string(len); \
+        memmove(String_val(dst), b, len);
 
 inline static char* strncat_safe(char* dest, char* src, int pos, int len)
 {
@@ -198,12 +199,9 @@ static void build_comments(vorbis_comment *vc,
 }
 
 
-static __inline__ value ocaml_vorbis_get_header_frame(value encoder, value title, value artist, value genre, value date, value album, value tracknum, value comment)
+static value ocaml_vorbis_get_header_frame(value encoder, value title, value artist, value genre, value date, value album, value tracknum, value comment)
 {
-  /* CAMLparam5(encoder, title, artist, genre, date);
-  CAMLxparam3(album, tracknum, comment); */
-  CAMLlocal1(s);
-
+  value s;
   char *ret = NULL;
   int ret_len = 0;
   myvorbis_encoder_t *ve = Encoder_val(encoder);
@@ -240,9 +238,9 @@ static __inline__ value ocaml_vorbis_get_header_frame(value encoder, value title
     vorbis_comment_clear(&comments);
   }
 
-  copy_buffer(ret, ret_len, s);
+  copy_buffer(s, ret, ret_len);
   free(ret);
-  /* CAMLreturn(s); */
+
   return s;
 }
 
@@ -361,20 +359,22 @@ CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value arti
 
 CAMLprim value ocaml_vorbis_create_encoder_byte(value* argv, int argc)
 {
-  return (ocaml_vorbis_create_encoder(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]));
+  return ocaml_vorbis_create_encoder(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]);
 }
 
 /* Encode audio from buf, return the available ogg/vorbis data */
 CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf, value _ofs, value _len)
 {
-  CAMLparam4(v_e, _buf, _ofs, _len);
-  CAMLlocal1(s);
+  CAMLparam2(v_e, _buf);
+  value s;
   myvorbis_encoder_t *ve = Encoder_val(v_e);
   int len = Int_val(_len);
+  int ofs = Int_val(_ofs);
   char* in = (char*)malloc(len*sizeof(char));
-  memcpy(in, String_val(_buf)+Int_val(_ofs), len*sizeof(char));
 
-  /* TODO: check that ofs + len <= length(buf) */
+  if (ofs + len > caml_string_length(_buf))
+    caml_invalid_argument("buffer too small (Vorbis.encode_buffer)");
+  memcpy(in, String_val(_buf)+ofs, len*sizeof(char));
 
   /* Ask vorbis the adress where to put the new data.
    * TODO some data can be lost when downsampling: with input at 22050Hz
@@ -471,7 +471,7 @@ CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf, value _ofs, val
   /* Back to the Caml world */
   caml_leave_blocking_section();
 
-  copy_buffer(ret, ret_len, s);
+  copy_buffer(s, ret, ret_len);
   free(ret);
   free(in);
 
@@ -512,14 +512,11 @@ static struct custom_operations decfile_ops =
 
 CAMLprim value ocaml_vorbis_open_dec_file(value fd, value params)
 {
-  CAMLparam2(fd, params);
-  CAMLlocal1(block);
+  CAMLparam1(params);
+  value block;
   int ret;
-
   FILE *in = NULL;
-  block = caml_alloc_custom(&decfile_ops, sizeof(myvorbis_dec_file_t*), 0, 1);
   myvorbis_dec_file_t *df = malloc(sizeof(myvorbis_dec_file_t));
-  Decfile_val(block) = df;
 
   if(!(in = fdopen(Int_val(fd), "rb")))
     caml_raise_constant(*caml_named_value("vorbis_exn_could_not_open_file"));
@@ -544,35 +541,52 @@ CAMLprim value ocaml_vorbis_open_dec_file(value fd, value params)
     raise_err(ret);
   }
 
+  block = caml_alloc_custom(&decfile_ops, sizeof(myvorbis_dec_file_t*), 0, 1);
+  Decfile_val(block) = df;
+
   CAMLreturn(block);
 }
 
 static size_t read_func_cb(void *ptr, size_t size, size_t nmemb, void *datasource)
 {
-  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
-  value buf = caml_callback(df->read_func, Val_int(size*nmemb));
-  int len = caml_string_length(buf);
+  myvorbis_dec_file_t *df = datasource;
+  value buf;
+  int len;
+  caml_leave_blocking_section();
+  buf = caml_callback(df->read_func, Val_int(size*nmemb));
+  len = caml_string_length(buf);
   memcpy(ptr, String_val(buf), len);
+  caml_enter_blocking_section();
   return len;
 }
 
 static int seek_func_cb(void *datasource, ogg_int64_t offset, int whence)
 {
-  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
-  return Int_val(caml_callback(df->seek_func, Val_int(offset)));
+  myvorbis_dec_file_t *df = datasource;
+  int ret;
+  caml_leave_blocking_section();
+  ret = Int_val(caml_callback(df->seek_func, Val_int(offset)));
+  caml_enter_blocking_section();
+  return ret;
 }
 
 static int close_func_cb(void *datasource)
 {
-  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
+  myvorbis_dec_file_t *df = datasource;
+  caml_leave_blocking_section();
   caml_callback(df->close_func, Val_unit);
+  caml_enter_blocking_section();
   return 0;
 }
 
 static long tell_func_cb(void *datasource)
 {
-  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
-  return Int_val(caml_callback(df->tell_func, Val_unit));
+  myvorbis_dec_file_t *df = datasource;
+  int ret;
+  caml_leave_blocking_section();
+  ret = Int_val(caml_callback(df->tell_func, Val_unit));
+  caml_enter_blocking_section();
+  return ret;
 }
 
 static ov_callbacks callbacks =
@@ -607,7 +621,9 @@ CAMLprim value ocaml_vorbis_open_dec_stream(value read_func, value seek_func, va
   caml_register_global_root(&df->close_func);
   df->tell_func = tell_func;
   caml_register_global_root(&df->tell_func);
+  caml_enter_blocking_section();
   ret = ov_open_callbacks(df, df->ovf, NULL, 0, callbacks);
+  caml_leave_blocking_section();
 
   if(ret < 0)
   {
@@ -621,7 +637,7 @@ CAMLprim value ocaml_vorbis_open_dec_stream(value read_func, value seek_func, va
 /* TODO: convert file format? */
 CAMLprim value ocaml_vorbis_decode(value d_f, value buf_, value ofs_, value len_)
 {
-  CAMLparam4(d_f, buf_, ofs_, len_);
+  CAMLparam2(d_f, buf_);
 
   myvorbis_dec_file_t *df = Decfile_val(d_f);
   int ret = 0;
@@ -631,7 +647,14 @@ CAMLprim value ocaml_vorbis_decode(value d_f, value buf_, value ofs_, value len_
 
   if (ofs + len > caml_string_length(buf_))
     caml_raise_constant(*caml_named_value("vorbis_exn_invalid_parameters"));
+
+  // We have to make sure that when a callback is called, the ocaml master lock
+  // has been released.  Callbacks are responsible for taking it back if they
+  // need to call ocaml code.
+  caml_enter_blocking_section();
   ret = ov_read(df->ovf, buf, len, df->big_endian, df->sample_size / 8, df->sign, &df->bitstream);
+  caml_leave_blocking_section();
+
   if (ret == 0)
   {
     free(buf);
@@ -661,7 +684,9 @@ CAMLprim value ocaml_vorbis_close_dec_file(value d_f)
     caml_remove_global_root(&df->close_func);
     caml_remove_global_root(&df->tell_func);
   }
+  caml_enter_blocking_section();
   ov_clear(df->ovf);
+  caml_leave_blocking_section();
   free(df->ovf);
 
   CAMLreturn(Val_unit);
@@ -669,39 +694,32 @@ CAMLprim value ocaml_vorbis_close_dec_file(value d_f)
 
 CAMLprim value ocaml_vorbis_utf8_set_charset(value charset)
 {
-  CAMLparam1(charset);
-
   convert_set_charset(String_val(charset));
-
-  CAMLreturn(Val_unit);
+  return Val_unit;
 }
 
 CAMLprim value ocaml_vorbis_utf8_decode(value string)
 {
-  CAMLparam1(string);
-  CAMLlocal1(ret);
-
+  value ret;
   char* utf8;
   if(utf8_decode(String_val(string), &utf8) >= 0)
   {
     ret = caml_copy_string(utf8);
     free(utf8);
-    CAMLreturn(ret);
+    return ret;
   }
   caml_raise_with_arg(*caml_named_value("vorbis_exn_utf8_failure"), string);
 }
 
 CAMLprim value ocaml_vorbis_utf8_encode(value string)
 {
-  CAMLparam1(string);
-  CAMLlocal1(ret);
-
+  value ret;
   char* nonutf8;
   if(utf8_encode(String_val(string), &nonutf8) >= 0)
   {
     ret = caml_copy_string(nonutf8);
     free(nonutf8);
-    CAMLreturn(ret);
+    return ret;
   }
   caml_raise_with_arg(*caml_named_value("vorbis_exn_utf8_failure"), string);
 }
@@ -772,28 +790,27 @@ CAMLprim value ocaml_vorbis_encoder_reset(value v_e, value title, value artist,
     size += (&s->og)->header_len + (&s->og)->body_len;
   }
 
-  copy_buffer(ret, size, ans);
+  copy_buffer(ans, ret, size);
   free(ret);
   CAMLreturn(ans);
 }
 
 value ocaml_vorbis_encoder_reset_byte(value *argv, int argc)
 {
-  return (ocaml_vorbis_encoder_reset(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]));
+  return ocaml_vorbis_encoder_reset(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]);
 }
 
-CAMLprim ocaml_vorbis_get_dec_file_bitstream(value d_f)
+CAMLprim value ocaml_vorbis_get_dec_file_bitstream(value d_f)
 {
-  CAMLparam1(d_f);
   myvorbis_dec_file_t *df = Decfile_val(d_f);
-  CAMLreturn(Val_int(df->bitstream));
+  return Val_int(df->bitstream);
 }
 
-CAMLprim ocaml_vorbis_get_dec_file_info(value d_f)
+CAMLprim value ocaml_vorbis_get_dec_file_info(value d_f)
 {
   CAMLparam1(d_f);
-  myvorbis_dec_file_t *df = Decfile_val(d_f);
   CAMLlocal1(ans);
+  myvorbis_dec_file_t *df = Decfile_val(d_f);
   vorbis_info *vi = df->ovf->vi;
 
   ans = caml_alloc_tuple(9);
@@ -811,21 +828,24 @@ CAMLprim ocaml_vorbis_get_dec_file_info(value d_f)
   CAMLreturn(ans);
 }
 
-CAMLprim ocaml_vorbis_get_dec_file_comments(value d_f, value link_)
+CAMLprim value ocaml_vorbis_get_dec_file_comments(value d_f, value link_)
 {
   CAMLparam2(d_f, link_);
+  CAMLlocal2(ans, cmts);
+
   myvorbis_dec_file_t *df;
   int link;
   int i;
   vorbis_comment *vc;
-  CAMLlocal2(ans, cmts);
 
   df = Decfile_val(d_f);
   if (Is_block(link_))
     link = Int_val(Field(link_, 0));
   else
     link = -1;
+  caml_enter_blocking_section();
   vc = ov_comment(df->ovf, link);
+  caml_leave_blocking_section();
 
   if (!vc)
   {
